/*    ********************************
      *      CONTR & BASIC           *
      *   UA6ANN  Alexander Kostyuk  *
      *   RV3BJ   Evgeny Fadeev      *
      ********************************
*/
/************************************************************************************
 $Workfile:  $
 $Header:    $
 $Log:       $
************************************************************************************/

#include "includes.h"

SERIAL_DCB serial_dcb[2];

UINT8 open_port(UINT8 port, INT32 baud_rate,UINT16 rxbufsize, UINT16 txbufsize)
{
UINT8 sreg;
SERIAL_DCB tmp_dcb;

if((sreg = SREG) & 0x80)
__disable_interrupt();

if(create_uart_DCB(&tmp_dcb,rxbufsize,rxbufsize))
  {
   tmp_dcb.BaudRate = baud_rate;
   set_uart_DCB (port, &tmp_dcb);
   return 1;
  }
if(sreg & 0x80)
__enable_interrupt();
return 0;
}

void close_port(UINT8 port)
{
UINT8 sreg;
UINT8 i,tmp;


if((sreg = SREG) & 0x80)
__disable_interrupt();

free(serial_dcb[port].rxbuff);
free(serial_dcb[port].txbuff);
fill_default_dcb(&(serial_dcb[port]));

if (port == 0)
       {
        UCSR0B = 0; // disable UART 0
        UCSR0A = (1<<TXC0); // disable UART 0
        for(i=0;i<10;i++) tmp = UDR0;
       }
else if (port == 1)
       {
        UCSR1B = 0; // disable UART 1
        UCSR1A = (1<<TXC1); // disable UART 0
        for(i=0;i<10;i++) tmp = UDR1;
       }
tmp=tmp; // to prevent compiler warnings
if(sreg & 0x80)
__enable_interrupt();
}



#pragma vector = USART0_RXC_vect
__interrupt void USART0_RX__isr(void)
{
  on_uart_rx_isr(0,UDR0);
}

#pragma vector = USART0_TXC_vect
__interrupt void USART0_TX_isr(void)
{
on_uart_tx_isr(0);
}

#pragma vector = USART1_RXC_vect
__interrupt void USART1_RX__isr(void)
{
  on_uart_rx_isr(1,UDR1);
}

#pragma vector = USART1_TXC_vect
__interrupt void USART1_TX_isr(void)
{
on_uart_tx_isr(1);
}





/********************************************************************************
Action:
Creates default Device Control Block, clears buffers pointers and buffers size variables
Sintax:
void fill_default_dcb(SERIAL_DCB *dcb);
serial_dcb = get_uart(port, rx_buf_size, tx_buf_size);
Inputs:
pointer to the DCB structure
Return: none
*********************************************************************************/
void fill_default_dcb (SERIAL_DCB *dcb)
{
memset((char*)dcb,0x00,sizeof(SERIAL_DCB));
dcb->BaudRate = 19200;
dcb->DataBits = DATA_8_BITS;
dcb->StopBits = ONE_STOP_BITS;
dcb->Parity = NOPARITY;
dcb->RTS_mode = RTS_TOGGLE;
dcb->TxDelay = 50;
dcb->SymTimeOut = 2;
dcb->RTS_ReleaseDelay = 2;
//  private part of structure
dcb->txbuff = 0;
dcb->rxbuff = 0;
//dcb->tx_wr_ndx = 0;
//dcb->tx_rd_ndx = 0;
//dcb->rx_wr_ndx = 0;
//dcb->rx_rd_ndx = 0;
dcb->rx_bsize = 0;
dcb->tx_bsize = 0;
//dcb->rx_cnt = 0;
//dcb->tx_cnt = 0;
dcb->tx_ready = 1;
}

/********************************************************************************
Action:
Creates default Device Control Block and allocates memory for buffers, sets
buffers size
Sintax:
INT8 create_uart_DCB(SERIAL_DCB *dcb,INT16 rxbufsize, INT16 txbufsize);
var = create_uart_DCB(dcb, rxbufsize, txbufsize);
Inputs:
pointer to the DCB structure
rx_buf_size and  tx_buf_size are size of buffers
Return: 1 - dcb created, buffers memory allocated
        0 - no memory to allocate
*********************************************************************************/
UINT8 create_uart_DCB (SERIAL_DCB *dcb, INT16 rxbufsize, INT16 txbufsize)
{
UINT8 *rxbuf,*txbuf;

fill_default_dcb(dcb);

if((rxbuf = malloc(rxbufsize))==0) return 0;
if((txbuf = malloc(txbufsize))==0)
         {
          free(rxbuf);
          return 0;
         }
fill_default_dcb(dcb);
dcb->txbuff = txbuf;
dcb->rxbuff = rxbuf;
dcb->rx_bsize = rxbufsize;
dcb->tx_bsize = txbufsize;

return 1;
}

/**************************************************************************
Action:
Sets up current port with settins in serial_dcb structure
Sintax:
SERIAL_DCB *serial_dcb;
set_uart(port, rx_buf_size, tx_buf_size);
Inputs:
port is number of serial ports 0..9
rx_buf_size and  tx_buf_size are size of buffers from 32 to 2048 bytes
Return: none
**************************************************************************/
void set_uart_DCB (UINT8 port, SERIAL_DCB  *dcb)
{
long div_const;
memcpy((char*)&(serial_dcb[port]),(char*)dcb,sizeof(SERIAL_DCB));
div_const = (FOSC/16/serial_dcb[port].BaudRate - 1);

if(port==0)
{
  UCSR0B = 0;         // disable UART
  UBRR0L =  div_const; // set Baud rate
  UBRR0H =  div_const >> 8; // set Baud rate
  UCSR0B = 0xD8;      // enable uart receive + transmit + rxc interrupt
}
else if(port==1)
{
  UCSR1B = 0;         // disable UART
  UBRR1L =  div_const; // set Baud rate
  UBRR1H =  div_const >> 8; // set Baud rate
  UCSR1B = 0xD8;      // enable uart receive + transmit + rxc interrupt
}


}
/**************************************************************************
Action:
Reads  data from a UART input buffering
Sintax:
var = uart_getchar(port);
Inputs: port  is one of serial ports 0..9
Return: value as shown in the table below
**************************************************************************/
INT16 uart_getchar (UINT8 port)
{
#define rx_rd_ndx     (serial_dcb[port].rx_rd_ndx)
#define rxbuff        (serial_dcb[port].rxbuff)
#define rx_bsize      (serial_dcb[port].rx_bsize)
#define rx_cnt        (serial_dcb[port].rx_cnt)


char data;

if(rx_bsize==0) return -1;

if(rx_cnt == 0)
     return -1;

data = rxbuff[rx_rd_ndx];
if (++rx_rd_ndx == rx_bsize) rx_rd_ndx = 0;
__disable_interrupt();
--rx_cnt;
__enable_interrupt();
return data;

#undef rx_rd_ndx
#undef rxbuff
#undef rx_bsize
#undef rx_cnt
}
/**************************************************************************
Action:
	Sends data through a UART output buffering
Sintax:
	var = uart_putchar(port, data);
Inputs: port  is on of serial ports 0..9, data - byte of data to be sent
Return: var is echoed data
**************************************************************************/
UINT8 uart_putchar (UINT8 port, UINT8 ch)
{

#define tx_wr_ndx     (serial_dcb[port].tx_wr_ndx)
#define txbuff        (serial_dcb[port].txbuff)
#define tx_bsize      (serial_dcb[port].tx_bsize)
#define tx_cnt        (serial_dcb[port].tx_cnt)

if(tx_bsize==0) return ch;

while (tx_cnt == tx_bsize);

__disable_interrupt();
if (tx_cnt || !serial_dcb[port].tx_ready)
  {
   txbuff[tx_wr_ndx] = ch;
   if (++tx_wr_ndx == tx_bsize) tx_wr_ndx = 0;
   ++tx_cnt;

  }
else
   UART_SEND_BYTE(port,ch);
__enable_interrupt();

return ch;

#undef tx_wr_ndx
#undef txbuff
#undef tx_bsize
#undef tx_cnt
}

/**************************************************************************
Action:
	Sends zero terminated string through an UART output buffering
Sintax:
	var = uart_putstr(port, string);
Inputs: port  is on of serial ports 0..9, data - zero terminated string
Return: var is number of sent bytes
**************************************************************************/
INT16 uart_putstr (UINT8 port, INT8 *data)
{
INT16 cnt = 0;
    while(*data)
      {
       uart_putchar(port,*data);
       cnt++;
       data++;
      }
return cnt;
}
/**************************************************************************
Action:
	Sends zero terminated string through a UART output buffering. String is located in the flash program memory.
Sintax:

	var = uart_putstr_P(port, msg_in_flash);

Inputs: port  is on of serial ports 0..9, msg_in_flash - zero terminated string in flash memory
Return: var is number of sent bytes
**************************************************************************/
INT16 uart_putstr_P (UINT8 port, INT8 __flash *data)
{

return 0;
}
/**************************************************************************
Action:
	Sends block of bytes through a UART output buffering.
Sintax:

	var = uart_putblock(port, data_block,len);

Inputs: port  is on of serial ports 0..9, data_block  is block of data, len is length of data block
Return: var is number of sent bytes
**************************************************************************/
INT16 uart_putblock (UINT8 port, INT8 *data, INT16 len)
{

return 0;
}

/********************************************************/
// UARTn Transmitter interrupt service routine
/********************************************************/
void on_uart_tx_isr(UINT8 port)
{
#define tx_rd_ndx     (serial_dcb[port].tx_rd_ndx)
#define txbuff        (serial_dcb[port].txbuff)
#define tx_bsize      (serial_dcb[port].tx_bsize)
#define tx_cnt        (serial_dcb[port].tx_cnt)
#define tx_error      (serial_dcb[port].tx_error)

serial_dcb[port].tx_ready = 1;

if (tx_cnt)
  {
   --tx_cnt;
   UART_SEND_BYTE(port,txbuff[tx_rd_ndx]);
   if (++tx_rd_ndx == tx_bsize) tx_rd_ndx = 0;
   };

#undef tx_rd_ndx
#undef txbuff
#undef tx_bsize
#undef tx_cnt
#undef tx_error

}


/********************************************************/
// UARTn Receiver interrupt service routine
/********************************************************/
void on_uart_rx_isr(UINT8 port, UINT8 data)
{
#define rx_wr_ndx     (serial_dcb[port].rx_wr_ndx)
#define rxbuff        (serial_dcb[port].rxbuff)
#define rx_bsize      (serial_dcb[port].rx_bsize)
#define rx_cnt        (serial_dcb[port].rx_cnt)
#define rx_ovrflow    (serial_dcb[port].rx_ovrflow)
   if(rx_bsize==0) return;
   rxbuff[rx_wr_ndx] = data;
   if (++rx_wr_ndx == rx_bsize) rx_wr_ndx = 0;
   if (++rx_cnt == rx_bsize)
      {
      rx_cnt = 0;
//      rx_ovrflow = 1;
      };

#undef rx_wr_ndx
#undef rxbuff
#undef rx_bsize
#undef rx_cnt
#undef rx_ovrflow



}


void UART_SEND_BYTE(UINT8 port, UINT8 c)
{
  serial_dcb[port].tx_ready = 0;
  if(port==0)
    UDR0 = c;
 else if(port==1)
    UDR1 = c;
}


