/*    ********************************
      *      CONTR & BASIC           *
      *   UA6ANN  Alexander Kostyuk  *
      *   RV3BJ   Evgeny Fadeev      *
      ********************************
*/
/************************************************************************************
 $Workfile:  $
 $Header:    $
 $Log:       $
************************************************************************************/
#include "includes.h"

int get_text_token(void);
char look_up_opr(char *s);
char a_to_hex(char ch);
short add_fix_label(short label,short prog_ndx);
short find_fix_label(short label);


/*
**********************************************
        Flash memory structure:
**********************************************

0x0000------------------+
|                       |
|      Firmware         |
|                       |
+-----------------------+
|     Label table       |
+-----------------------+
|                       |
|                       |
|     Internal Basic    |
| program representation|
|                       |
|                       |
+-----------------------+
|    INLINE CODE area   |
+-----------------------+
|    Flash utilities    |
|     in BOOT area      |
0xXFFF------------------+

Editor text:
PRINT "INLINE TEST"
A=85
INLINE 58629,47874,38152
PRINT "OK"
END

--------------------------------
Internal format in flash memory:
--------------------------------
04 01 06 0c 49 4e 4c 49 4e 45 20 54 45 53 54 00 0d 
02 00 3d 03 00 55 0d 
04 2b 03 fa 00 0d 
04 01 06 03 4F 4B 00 0d 
04 10 0d 00 

----------------
INLINE CODE area 
----------------
FA00 05 e5 02 bb 08 95 ff ff ff ff ff ff ff ff ff ff

-----------------------------------------
Detail explain converted program lines
-----------------------------------------
1-st line
04  - COMMAND 
01  - PRINT
06  - QUOTE
0c  - Symbol counter
49 4e 4c 49 4e 45 20 54 45 53 54 00 - Zero terminated string "INLINE TEST"
0d - EOL
-------------------
2-nd line
02      - VARIABLE
00      - VARIABLE A
3d      = operator
03      - NUMBER
00 55   - value 85
0d      EOL
-------------------
3-rd line
04       - COMMAND
2b       - INLINE
03       - NUMBER
fa 00    - Start address in the INLINE CODE area filled at the downloading stage
           The procedure in the INLINE CODE area must be formed as a subroutine.
0d       - EOL
-------------------
4-th line
04         - COMMAND
01         - PRINT 
06         - QUOTE
03         - Symbol counter
4F 4B 00   - Zero terminated string "OK"
0d         - EOL
--------------------
5-th line
04         - COMMAND  
10         - END
0d         - EOL
00         - EOF

*/
/*******************************************************************
Converts input line to the internal format to put it into the flash 
program memory at the downloading stage
Inputs:
prog_ndx - current index in the flash memory
ibuf     - pointer to the input line
wp       - pointer to the output buffer

Returns pointer to the converted line
*******************************************************************/
unsigned short prog_ndx;
char *convert_line_to_intformat(unsigned short prog_ndx, char *ibuf, char *wp)
{
char use_label_flag = 0;
int i;
char *w;
short tmp16;
short lblndx;
short sign;
unsigned char token_ndx = 0;
char *p = ibuf;
bas_prog = wp;
do
 {
 token_ndx++;
 get_text_token();
   switch(bas_token_type)
   {
     case DELIMITER:
       if(*bas_token_ext)
           *p++ = *bas_token_ext;
       else
          {
           *p = 0;
           goto lab002;
          }
     break;
     case VARIABLE:
       *p++ = VARIABLE;
       *p++ = toupper(*bas_token_ext)-'A';
     break;
     case OPERATOR:
       *p++ = bas_tok;
     break;
     case NUMBER:
          tmp16 = bas_int_number;
       if(use_label_flag)
       {
         use_label_flag = 0;
         if((lblndx = find_fix_label(tmp16)) == -1)
          {
            if( (lblndx = add_fix_label(tmp16,-1)) == -1)
                                           serror(5,0); // full
          }
         *p++ = NUMBER;
           if(lblndx != -1)
            {
             *p++ = lblndx >> 8;
             *p++ = lblndx;
            }
            else
             {
             *p++ = 0x00;
             *p++ = 0x00;

             }
       }
       else if(token_ndx == 1)
       {
         if((lblndx = find_fix_label(tmp16)) != -1)
          {
            if(label_table[lblndx].prog_ndx != -1)
                     serror(6,0); // duplicate
             else
               label_table[lblndx].prog_ndx = (prog_ndx + (p - ibuf));

          }
          else
           {
              if(  add_fix_label(tmp16,prog_ndx + (p - ibuf)) == -1)
                 serror(5,0); // full
           }

        }
        else
        {
          *p++ = NUMBER;
          *p++ = tmp16 >> 8;
          *p++ = tmp16;
        }
     break;
     case FUNC:
          *p++ = FUNC;
          *p++ = bas_tok;
       break;
     case COMMAND:
          *p++ = COMMAND;
          *p++ = bas_tok;
         if((bas_tok == GOTO)||(bas_tok == GOSUB)||(bas_tok == RESTORE))
                                 use_label_flag = 1;

         if(bas_tok == REM)
         {
          i = 1;
          *p++ = STRING;
          w = p++;
          while(bas_prog[i]>= 0x20)
           *p++ = bas_prog[i++];

           *p++ = 0x00;
           *p++ = 0x0D;
          *w = i + 1;
          return p;
         }

#ifdef MAXI_SYSTEM
         if(bas_tok == INLINE)
         {
          w=p++;
          tmp16 = START_OF_INLINE_CODE + (code_buf_ptr - code_buffer);
          *p++ = tmp16 >> 8;
          *p++ = tmp16;
          i=0;
          sign = 1;
          do{
          get_text_token();
            if(bas_token_type == DELIMITER)
             {
               if (*bas_token_ext=='+') sign = 1;
               else if(*bas_token_ext=='-') sign = -1;
                else if(!((*bas_token_ext == ',')||(bas_tok == EOL)))serror(0,0);
             } else
            if(bas_token_type == NUMBER)
              {
               i++;
                tmp16 = sign * bas_int_number;
                sign = 1;
                *code_buf_ptr++ = tmp16;
                *code_buf_ptr++ = tmp16 >> 8;
              } else serror(0,0);
          } while(!((bas_token_type == DELIMITER)&&(bas_tok == EOL)));
          *w = i;
          *p++ = 0x0D;

         }
#endif /*MAXI_SYSTEM*/
         if(bas_tok == DATA)
         {
          sign = 1;
          do{
          get_text_token();
            if(bas_token_type == DELIMITER)
             {
               if (*bas_token_ext=='+') sign = 1;
               else if(*bas_token_ext=='-') sign = -1;
                else if(!((*bas_token_ext == ',')||(bas_tok == EOL)))serror(0,0);
                if(*bas_token_ext == ',')
                  *p++ = ',';
             } else
            if(bas_token_type == NUMBER)
              {
                tmp16 = sign * bas_int_number;
                sign = 1;
                *p++ = NUMBER;
                *p++ = tmp16 >> 8;
                *p++ = tmp16;
              } else serror(0,0);
          } while(!((bas_token_type == DELIMITER)&&(bas_tok == EOL)));
          *p++ = 0x0D;
         }
#ifdef MAXI_SYSTEM
         if(bas_tok == INTERRUPT)
         {
          w=p++;
          get_text_token();
          if(bas_token_type == NUMBER)
             tmp16 = bas_int_number;
          else
             tmp16 = 0;
          *p++ = tmp16; // vector address, low byte
          tmp16 = START_OF_INLINE_CODE + (code_buf_ptr - code_buffer);
          *p++ = tmp16 >> 8;
          *p++ = tmp16;
          i=0;
         sign = 1;
         do{
          get_text_token();
            if(bas_token_type == DELIMITER)
             {
               if (*bas_token_ext=='+') sign = 1;
               else if(*bas_token_ext=='-') sign = -1;
                else if(!((*bas_token_ext == ',')||(bas_tok == EOL)))serror(0,0);
             } else
            if(bas_token_type == NUMBER)
              {
               i++;
                tmp16 = sign * bas_int_number;
                sign = 1;
                *code_buf_ptr++ = tmp16;
                *code_buf_ptr++ = tmp16 >> 8;
              }
              else serror(0,0);
          } while(!((bas_token_type == DELIMITER)&&(bas_tok == EOL)));
          *w = i;
          *p++ = 0x0D;
         }
#endif /*MAXI_SYSTEM*/

     break;
      case QUOTE:
       *p++ = QUOTE;
       w = p++;
       *w = 0;
        i = 0;
        while(bas_token_ext[i])
           *p++ = bas_token_ext[i++];

           *p++ = 0x00;
           *w = i + 1;
     break;
     default:
        bas_token_ext++;
        serror(0,0);
        return p;
   }
 } while((bas_tok!=FINISHED)||(DELIMITER != bas_token_type));
lab002:
return p;
}

short find_fix_label(short label)
{
char i;

  for(i=0;i<NUM_LAB;i++)
  {
   if(label_table[i].label_value == label)
      return i;
  }
return -1;
}


short add_fix_label(short label,short prog_ndx)
{
char i;

  for(i=0;i<NUM_LAB;i++)
  {
   if((label_table[i].prog_ndx == -1) && (label_table[i].label_value == -1))
   {
   label_table[i].prog_ndx = prog_ndx;
   label_table[i].label_value = label;
   return i;
   }
  }
return -1;
}


char a_to_hex(char ch)
{

 if(ch<='9') return ch & 0x0F;
  else
   return toupper(ch) -'A' + 10;
}

char look_up_opr(char *s)
{
  register int i;
  char *p;

  /* convert to lowercase */
  p = s;
  while(*p){ *p = toupper(*p); p++; }

  /* see if token is in table */
  for(i=0; *opers_table[i].command; i++)
      if(!strcmp_P(s, opers_table[i].command))
                     return opers_table[i].tok;
  return 0; /* unknown command */
}



/**************************************************************
                        Get a token. 
***************************************************************/
int get_text_token(void)
{
  char *temp;
unsigned short tmpu16;
unsigned short mult;
char i;
  bas_token_type=0; bas_tok=0;
  temp=bas_token_ext;
  if(*bas_prog=='\0') { /* end of file */
    *bas_token_ext=0;
    bas_tok = FINISHED;
    return(bas_token_type=DELIMITER);
  }

  while(iswhite(*bas_prog)) ++bas_prog;  /* skip over white space */

  if(*bas_prog=='\r') { /* crlf */
    ++bas_prog; ++bas_prog;
    bas_tok = EOL;
    bas_token_ext[0]='\r';
    bas_token_ext[1]='\n'; bas_token_ext[2]=0;

    return (bas_token_type = DELIMITER);
  }

  if(strchr("+-*/=;(),><:", *bas_prog)){ /* delimiter */
    *temp=*bas_prog;
    bas_prog++; /* advance to next position */
    temp++;
    *temp=0;
    return (bas_token_type=DELIMITER);
  }

  if(*bas_prog=='"') { /* quoted string */
    bas_prog++;
    while(*bas_prog!='"'&& *bas_prog!='\r' && *bas_prog!=0x00) *temp++=*bas_prog++;
    if((*bas_prog=='\r') || (*bas_prog==0x00))
         {
          serror(1,0);
          return(bas_token_type=QUOTE);
         }
      bas_prog++;
   *temp=0;
    return(bas_token_type=QUOTE);
  }

  if(isdigit(*bas_prog)) { /* number */
    while(!isdelim(*bas_prog)) *temp++=*bas_prog++;
    *temp = '\0';
    bas_int_number = atoi(bas_token_ext);
    return(bas_token_type = NUMBER);
  }

  if( (bas_prog[0]=='&') && ((bas_prog[1]=='H') || (bas_prog[1]=='h')))
  {
    bas_prog +=2;
     while(strchr("0123456789ABCDEFabcdef",*bas_prog)) *temp++=a_to_hex(*bas_prog++);
     i = temp - bas_token_ext;
     mult = 1;
     bas_int_number = 0;
     for(;i>0;i--)
      {
       temp--;
       tmpu16 = *temp * mult;
       bas_int_number += tmpu16;
       mult *= 16;
      }
    return(bas_token_type = NUMBER);
  }

  if( (bas_prog[0]=='&') && ((bas_prog[1]=='B') || (bas_prog[1]=='b')))
  {
    bas_prog +=2;
     while(strchr("01",*bas_prog)) *temp++=*bas_prog++;
     i = temp - bas_token_ext;
     mult = 1;
     bas_int_number = 0;
     for(;i>0;i--)
      {
       temp--;
       tmpu16 = (*temp - '0') * mult;
       bas_int_number += tmpu16;
       mult *= 2;
      }
    return(bas_token_type = NUMBER);
  }


  if(isalpha(*bas_prog)) { /* var or command */
    while(!isdelim(*bas_prog)) *temp++=*bas_prog++;
    bas_token_type=STRING;
  }

  *temp = '\0';

  /* see if a string is a command or a variable */
  if(bas_token_type==STRING) {
    bas_tok=look_up(bas_token_ext); /* convert to internal rep */
    if(bas_tok) bas_token_type = COMMAND;
    else
    {
     bas_tok=look_up_fun(bas_token_ext); /* convert to internal rep */
     if(bas_tok) bas_token_type = FUNC;
     else
       {
         bas_tok=look_up_opr(bas_token_ext); /* convert to internal rep */
         if(bas_tok) bas_token_type = OPERATOR;
          else
             bas_token_type = VARIABLE;
       }
   }

  }
  return bas_token_type;
}
